import dayjs from 'dayjs'
import { getActualPoint, getAbsoluteCoordinate, getInitialPoint, getHtmlToText } from '../utils';

export default class Freetext {
  
  constructor ({
    tool,
    color,
    svgElement,
    layer,
    container,
    viewport,
    page,
    eventBus
  }) {
    this.layer = layer
    this._tool = tool
    this.color = color
    this.svgElement = svgElement
    this.container = container
    this.viewport = viewport
    this.page = page
    this.eventBus = eventBus

    this.freetextElement = null
    this.start = null
    this.end = null

    this.onClick = this.handleClick.bind(this)
    this.onBlur = this.handleBlur.bind(this)
    this.init()
  }

  get tool () {
    return this._tool
  }

  set tool (toolType) {
    if (toolType === this._tool) return
    if (!toolType) {
      this.reset()
      this.svgElement.style.cursor = 'default'

      this.container.removeEventListener('click', this.onClick)
    }
    this._tool = toolType
  }

  reset () {
    const freetextElement = this.freetextElement
    this.start = null
    this.end = null

    this.container.removeEventListener('click', this.onClick)
    if (freetextElement) {
      freetextElement.remove()
      freetextElement.removeEventListener('blur', this.onBlur)
      this.freetextElement = null
    }
  }

  init () {
    this.container.addEventListener('click', this.onClick)
    this.svgElement.style.cursor = 'crosshair'
  }

  getFreetextRectPoint () {
    const start = this.start
    const end = this.end
    const startX = start.x
    const startY = end.y
    const endX = end.x
    const endY = start.y
    return {
      start: {
        x: startX,
        y: startY
      },
      end: {
        x: endX,
        y: endY
      }
    }
  }

  getFreetextRectString () {
    const viewport = this.viewport
    const { scale } = this.viewport
    const freetextRectPoint = this.getFreetextRectPoint()
    const actualStartPoint = getInitialPoint(freetextRectPoint.start, viewport, scale)
    const actualEndPoint = getInitialPoint(freetextRectPoint.end, viewport, scale)
    return `${actualStartPoint.x},${actualStartPoint.y},${actualEndPoint.x},${actualEndPoint.y}`
  }

  handleFreetext () {
    const freetextElement = this.freetextElement
    if (freetextElement.innerText) {
      const rect = freetextElement.getBoundingClientRect()
      this.end = {
        x: this.start.x + rect.width,
        y: this.start.y + rect.height
      }
      const freetextRectString = this.getFreetextRectString()

      const text = getHtmlToText(freetextElement)

      const annotation = {
        operate: "add-annot",
        type: 'freetext',
        page: this.page,
        textColor: '#000000',
        fillColor: '#FFFFFF',
        fontSize: 16,
        fillTransparency: 0,
        color: 'transparent',
        fontName: 'Helvetica',
        transparency: 1,
        alignment: 0,
        width: 0,
        fillTransparency: 0,
        content: text,
        rect: freetextRectString,
        modified: dayjs(new Date()).format('YYYYMMDDHHmmss')
      }

      this.eventBus.dispatch('annotationChange', {
        type: 'add',
        annotation
      })
      this.layer.renderAnnotation(annotation, true)
    }

    freetextElement.remove()
    freetextElement.removeEventListener('blur', this.onBlur)
    this.freetextElement = null
  }

  handleBlur () {
    this.handleFreetext()
  }

  handleClick (event) {
    if (this.layer.selectedElementName) return
    const freetextElement = this.freetextElement
    if (freetextElement && event.target === freetextElement) return

    if (!freetextElement && !this.layer.selectedElementName) {
      this.renderFreetext(event)
    }
  }

  setCss (ele, cssText) {
    if (!ele) return
    if (cssText) {
      for (let key in cssText) {
        ele.style[key] = cssText[key]
      }
    }
  }

  renderFreetext (event) {
    const start = getAbsoluteCoordinate(this.container, event)
    this.start = start
    if (!this.freetextElement) {
      const { width, height } = this.viewport
      const maxWidth = width - this.start.x
      const maxHeight = height - this.start.y

      let freetextElement = document.createElement('div')
      freetextElement.setAttribute('role', 'textbox')
      freetextElement.setAttribute('contenteditable', true)
      freetextElement.setAttribute('spellcheck', false)
      this.setCss(freetextElement, {
        position: 'absolute',
        left: start.x - 4 + 'px',
        top: start.y - 2 + 'px',
        maxWidth: maxWidth+ 'px',
        maxHeight: maxHeight + 'px',
        minWidth: '20px',
        fontSize: '16px',
        textAlign: 'left',
        whiteSpace: 'pre-wrap',
        color: '#000',
        lineHeight: '1.2em',
        padding: '2px 4px',
        border: '1px solid #1460F3',
        overflow: 'auto'
      })
      freetextElement.className = 'freetext'
      freetextElement.addEventListener('blur', this.onBlur)

      this.freetextElement = freetextElement
      this.container.append(freetextElement)
      freetextElement.focus()
    }
  }
}
