import { getStroke } from "perfect-freehand";
import {
  getAbsoluteCoordinate,
  createSvg,
  getInitialPoint,
  getSvgPathFromStroke
} from '../utils';

export default class PaintShape {
  constructor ({
    tool,
    color,
    svgElement,
    layer,
    container,
    viewport,
    page,
    eventBus
  }) {
    this.layer = layer
    this._tool = tool
    this.color = color
    this.svgElement = svgElement
    this.container = container
    this.viewport = viewport
    this.page = page
    this.eventBus = eventBus
    this.path = []

    this.initStartPoint = null
    this.initEndPoint = null
    this.shapeEle = null

    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)

    this.init()
  }

  init () {
    this.container.addEventListener('mousedown', this.onMousedown)
    this.container.addEventListener('touchstart', this.onMousedown)
    this.svgElement.style.cursor = 'crosshair'
    this.container.style.touchAction = 'none'
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
  }

  get tool () {
    return this._tool
  }

  set tool (toolType) {
    if (toolType === this._tool) return
    if (!toolType) {
      this.reset()
      this.svgElement.style.cursor = 'default'
      this.container.style.touchAction = ''

      this.container.removeEventListener('mousedown', this.onMousedown)
      this.container.removeEventListener('touchstart', this.onMousedown)
    }
    this._tool = toolType
  }

  reset () {
    this.initStartPoint = null
    this.initEndPoint = null
    this.shapeEle = null
    this.svgElement.innerHTML = ''
    this.path = []
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)

    this.container.removeEventListener('mousedown', this.onMousedown)
    this.container.removeEventListener('touchstart', this.onMousedown)
  }

  checkSvgChild (tool) {
    switch (tool) {
      case 'square':
        return 'rect';
      case 'circle':
        return 'ellipse';
      case 'ink':
        return 'path';
      default:
        return 'line';
    }
  }

  handleMouseDown (event) {
    if (this.layer.selectedElementName) return
    const { x, y } = getAbsoluteCoordinate(this.container, event)
    this.initStartPoint = {
      x,
      y
    }

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
  }

  handleMouseUp () {
    let initStartPoint = this.initStartPoint
    let initEndPoint = this.initEndPoint
    if (initEndPoint && initStartPoint.x !== initEndPoint.x && initStartPoint.y !== initEndPoint.y) {
      let start = getInitialPoint(initStartPoint, this.viewport, this.viewport.scale)
      let end = getInitialPoint(initEndPoint, this.viewport, this.viewport.scale)
      
      let linePoints = start.x + ',' + start.y + ',' + end.x + ',' + end.y

      const annotationData = {
        operate: "add-annot",
        type: this.tool === 'arrow' ? 'line' : this.tool,
        page: this.page,
        lineWidth: 2,
        transparency: 0.8,
        borderColor: this.color,
        linePoints,
        arrow: this.tool === 'arrow',
      }

      if (this.tool === 'circle' || this.tool === 'square') {
        const newStart = {
          x: Math.min(start.x, end.x),
          y: Math.min(start.y, end.y)
        }
        const newEnd = {
          x: Math.max(start.x, end.x),
          y: Math.max(start.y, end.y)
        }
        linePoints = newStart.x + ',' + newStart.y + ',' + newEnd.x + ',' + newEnd.y
        annotationData.rect = linePoints
        annotationData.linePoints = linePoints
      } else if (this.tool === 'ink') {
        const { path, rect } = this.formatPath()

        annotationData.rect = rect
        annotationData.inklist = path
      } else if (this.tool === 'line' || this.tool === 'arrow') {
        annotationData.rect = linePoints
      }

      this.eventBus.dispatch('annotationChange', {
        type: 'add',
        annotation: annotationData
      })
      this.layer.renderAnnotation(annotationData)

    }
    // this.reset()
    
    this.initStartPoint = null
    this.initEndPoint = null
    this.shapeEle = null
    this.svgElement.innerHTML = ''
    this.path = []
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
  }

  handleMouseMove (event) {
    let { x, y } = getAbsoluteCoordinate(this.container, event)
    const { width, height } = this.viewport
    const scale = this.viewport.scale
    x = Math.min(width, x)
    x = Math.max(0, x)
    y = Math.min(height, y)
    y = Math.max(0, y)
    this.initEndPoint = {
      x,
      y
    }
    if (this.initStartPoint.x !== this.initEndPoint.x && this.initStartPoint.y !== this.initEndPoint.y) {
      if (!this.shapeEle) {
        const shapeEle = createSvg(
          this.checkSvgChild(this.tool),
          {
            'fill': 'none',
            'stroke': this.color, 
            'stroke-width': 2 * scale,
            'stroke-opacity': 0.8
          }
        )
        if (this.tool === 'arrow') {
          const ele = document.createElementNS(
            'http://www.w3.org/2000/svg',
            'defs'
          )

          ele.innerHTML = `<marker
            id="arrow"
            markerWidth=2
            markerHeight=3
            refX=3
            refY=2
            orient="auto"
            markerUnits="strokeWidth"
          >
            <polyline
              points="0.25,0.5 3.25,2 0.25,4"
              stroke=${this.color}
              stroke-width=1
              fill="none"
              stroke-dasharray=100
              stroke-opacity=0.8
            />
          </marker>`;
          shapeEle.setAttribute('marker-end', 'url(#arrow)');
          this.svgElement.appendChild(ele);
        }
        this.shapeEle = shapeEle
        this.svgElement.append(this.shapeEle)
      } else {
        if (event.shiftKey && (this.tool === 'circle' || this.tool === 'square')) {
          const { initStartPoint, initEndPoint } = this
          const distanceX = Math.abs(initEndPoint.x - initStartPoint.x)
          const distanceY = Math.abs(initEndPoint.y - initStartPoint.y)
          const distance = Math.min(distanceX, distanceY)
          if (initStartPoint.x < initEndPoint.x) {
            initEndPoint.x = initStartPoint.x + distance
          } else {
            initEndPoint.x = initStartPoint.x - distance
          }
          if (initStartPoint.y < initEndPoint.y) {
            initEndPoint.y = initStartPoint.y + distance
          } else {
            initEndPoint.y = initStartPoint.y - distance
          }
        }
        const rect = this.calculate(this.initStartPoint, this.initEndPoint)

        const shapeEle = this.shapeEle
        if (this.tool === 'circle') {
          shapeEle.setAttribute('cx', rect.left + rect.width / 2,)
          shapeEle.setAttribute('cy', rect.top + rect.height / 2,)
          shapeEle.setAttribute('rx', (rect.width - 1) / 2,)
          shapeEle.setAttribute('ry', (rect.height - 1) / 2,)
        } else if (this.tool === 'square') {
          shapeEle.setAttribute('x', rect.left + 1 / 2,)
          shapeEle.setAttribute('y', rect.top - 1 / 2,)
          shapeEle.setAttribute('width', `${Math.abs(rect.width - 1)}px`,)
          shapeEle.setAttribute('height', `${Math.abs(rect.height - 1)}px`,)
        } else if (this.tool === 'line' || this.tool === 'arrow') {
          shapeEle.setAttribute('x1', this.initStartPoint.x);
          shapeEle.setAttribute('y1', this.initStartPoint.y);
          shapeEle.setAttribute('x2', this.initEndPoint.x);
          shapeEle.setAttribute('y2', this.initEndPoint.y);
        } else if (this.tool === 'ink') {
          this.path.push([x, y])
          const d = this.getFreeDrawSvgPath()
          shapeEle.setAttribute('d', d)
        }
      }
    }
  }

  getFreeDrawSvgPath () {
    const options = {
      simulatePressure: false,
      size: 1,
      thinning: 0,
      smoothing: 0.3,
      streamline: 0.5,
      easing: (t) => Math.sin((t * Math.PI) / 2),
      last: false
    }
    return getSvgPathFromStroke(getStroke(this.path, options))
  }

  handlePoint (point) {
    const newPoint = getInitialPoint(
      {
        x: point[0],
        y: point[1]
      },
      this.viewport,
      this.viewport.scale
    )
    return {
      x: newPoint.x,
      y: newPoint.y
    }
  }

  getRect () {
    const xArray = [];
    const yArray = [];
    this.path.forEach((point) => {
      xArray.push(point[0]);
      yArray.push(point[1]);
    });
    const top = Math.min(...yArray);
    const left = Math.min(...xArray);
    const bottom = Math.max(...yArray);
    const right = Math.max(...xArray);
    this.path = this.path.map((ele) => {
      let x = ele[0] - left
      let y = ele[1] - top
      return [x, y]
    })
    return {
      left,
      top,
      bottom,
      right,
      width: right - left,
      height: bottom - top
    }
  }

  formatPath () {
    let path = ''
    const length = this.path.length
    
    const xArray = [];
    const yArray = [];

    for (let i = 0; i < length; i = i+2) {
      const point = this.handlePoint(this.path[i])

      xArray.push(point.x);
      yArray.push(point.y);

      path+=(point.x.toFixed(2) + ',' + point.y.toFixed(2))
      if (i + 1 < length) {

        const point = this.handlePoint(this.path[i])

        xArray.push(point.x);
        yArray.push(point.y);

        path+=(';' + point.x.toFixed(2) + ',' + point.y.toFixed(2))
        if (i + 2 < length) {
          path += ';'
        }
      }
    }

    const left = Math.min(...xArray);
    const bottom = Math.min(...yArray);
    const right = Math.max(...xArray);
    const top = Math.max(...yArray);
    const width = right - left
    const height = top - bottom

    return {
      path,
      rect: `${left},${bottom},${right},${top}`
    }
  }
  
  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });
    const actualbdwidth = 1 * this.viewport.scale

    return {
      top: initRect.top - actualbdwidth,
      left: initRect.left - actualbdwidth,
      width: initRect.width + actualbdwidth * 2,
      height: initRect.height + actualbdwidth * 2,
    }
  }

}
